#include <stdio.h>
#include <stdlib.h>
#include <xmms/plugin.h>
#include <string.h>
#include <math.h>

#include <xmms/util.h>
#include <pthread.h>

// following also includes some X stuff we need
#include "getroot.c"

// some defines

/* following can be modified at your pleasure */
#define POSX 10
#define POSY 656
#define HEIGHT 100 // maximum height/amplitude of a bar
#define CUTOFF 180 // frequencies are represented by 256 values
// we're cutting off high frequencies by only showing 0 to CUTOFF
#define DIV 4 // we have CUTOFF sources, every bar represents DIV sources
#define SPACE 1 // space between bars
#define WIDTH 7 // width of a bar
// set the following to 0 to disable shadows
#define SHADOW 1 // offset of shadow in pixels
#define ORIENTATION 0 // 0 = bottom->up, 1 = top->down, 2 = left->right, 3 = right->left

// set the following to HEIGHT to disable falloff
#define FALLOFF 5 // how many pixels the bars should falloff every frame

// set the following to 0 to disable peaks
#define PEAK	4 // how many pixels the peaks should falloff every frame
#define PEAKSTEP	5 // how many frames should the peak resist the falloff

#define FPS	30 // how many frames per second should be drawn

// color of the bars (0 - 255):
#define RED0 163
#define GREEN0 196
#define BLUE0 34
#define RED1 184
#define GREEN1 221
#define BLUE1 39
#define RED2 205
#define GREEN2 246
#define BLUE2 43
#define RED3 230
#define GREEN3 255
#define BLUE3 100
// same for the peak
#define REDP 255
#define GREENP 255
#define BLUEP 255
// same for the shadow
#define REDS 0
#define GREENS 40
#define BLUES 80

/* Linearity of the amplitude scale (0.5 for linear, keep in [0.1, 0.9]) */
#define LINEARITY 0.33

#define DEBUG 0 // print out debug messages

// Forward declarations
static void rootvis_init(void);
static void rootvis_cleanup(void);
static void rootvis_about(void);
static void rootvis_configure(void);
static void rootvis_playback_start(void);
static void rootvis_playback_stop(void);
static void rootvis_render_freq(gint16 freq_data[2][256]);

// Callback functions
VisPlugin rootvis_vtable = {
	0, // Handle, filled in by xmms
	0, // Filename, filled in by xmms

	0,                     // Session ID
	"Root Spectrum Analyzer 0.0.4",  // description

	0, // # of PCM channels for render_pcm()
	1, // # of freq channels wanted for render_freq()    == MONO?

	rootvis_init,           // Called when plugin is enabled
	rootvis_cleanup,        // Called when plugin is disabled
	NULL,//rootvis_about,          // Show the about box
	NULL,//rootvis_configure,      // Show the configure box
	0,                     // Called to disable plugin, filled in by xmms
	rootvis_playback_start, // Called when playback starts
	rootvis_playback_stop,  // Called when playback stops
	0,                     // Render the PCM data, must return quickly
	rootvis_render_freq     // Render the freq data, must return quickly
};

// XMMS entry point
VisPlugin *get_vplugin_info(void) {
	return &rootvis_vtable;
}


// Some evil global variables
pthread_t work_thread;
pthread_mutex_t mutex1 = PTHREAD_MUTEX_INITIALIZER;
enum {GO, STOP} thread_control;

gint16 shared_freq_data[2][256];

// X related
Display *display;
int screen;
Window rootWin, Parent;
GC gc;
char *display_name = NULL;

// Some helper stuff

void clean_data(void) {
	pthread_mutex_lock(&mutex1);
	memset(shared_freq_data, 0, sizeof(gint16) * 2 * 256);
	pthread_mutex_unlock(&mutex1);
}

void print_status(char msg[]) {
	if (DEBUG == 1) printf(">> rootvis >> %s\n", msg); // for debug purposes, but doesn't tell much anyway
}

void error_exit(char msg[]) {
	printf("*** ERROR (rootvis): %s\n", msg);
	rootvis_vtable.disable_plugin(&rootvis_vtable);
}

void initialize_X(void) {
	// XGCValues values;
	// unsigned long valuemask;
	print_status("Opening X Display");
	display = XOpenDisplay(display_name);
	if (display == NULL) {
		if (display_name == NULL) display_name = getenv("DISPLAY");
		fprintf(stderr, "cannot connect to X server %s\n",
			display_name ? display_name : "(default)");
		error_exit("Connecting to X server failed");
	}
	print_status("Getting screen and window");
	screen = DefaultScreen(display);
	rootWin = ToonGetRootWindow(display, screen, &Parent);

	/* these variables are used to specify various attributes for the GC. */
	/* initial values for the GC. */
	//values.cap_style = CapButt;
	//values.join_style = JoinBevel;
	/* which values in 'values' to check when creating the GC. */
	// valuemask = NULL;
	print_status("Creating Graphical Context");
	gc = XCreateGC(display, rootWin, 0, NULL);
	if (gc < 0) {
		error_exit("Creating Graphical Context failed");
	}
	print_status("Setting Line Attributes");
	XSetLineAttributes(display, gc, 1, LineSolid, CapButt, JoinBevel);
}

long get_color(char red, char green, char blue) {
	Colormap def_colormap;
 	XColor dacolor;
	dacolor.red = red * 256;
	dacolor.green = green * 256;
	dacolor.blue = blue * 256;
	print_status("Getting Colormap");
	def_colormap = DefaultColormap(display, DefaultScreen(display));
	print_status("Allocationg color");
	if (XAllocColor(display, def_colormap, &dacolor) == 0) {
		error_exit("Color allocation failed");
	}
	return dacolor.pixel;
}

void draw_bar(int i, int color, int peakcolor, int shcolor, unsigned short level, unsigned short oldlevel, unsigned short peak, unsigned short oldpeak) {
	
	/* to make following cleaner, we work with redundant helper variables
	   this also avoids some calculations */
	register int a, b, c, d;
	int drawpeak = 0;
	
	if ((PEAK > 0)&&(peak != oldpeak)) drawpeak = 1;
	
	if (ORIENTATION < 2) { a = POSX + i*(WIDTH+SHADOW+SPACE); c = WIDTH; }
	else	{ b = POSY + (CUTOFF/DIV - i)*(WIDTH+SHADOW+SPACE); d = WIDTH; }
	
	if (level > oldlevel) {
		if ((PEAK > 0)&&(peak == level)) level--;
		if (ORIENTATION == 0) {	b = POSY + HEIGHT - level; d = level; }
		else if (ORIENTATION == 1) { b = POSY; d = level; }
		else if (ORIENTATION == 2) { a = POSX; c = level; }
		else	{ a = POSX + HEIGHT - level; c = level; }
		XSetForeground(display, gc, shcolor);
		XFillRectangle(display, rootWin, gc, a+SHADOW, b+SHADOW, c, d);
		XSetForeground(display, gc, color);
		XFillRectangle(display, rootWin, gc, a, b, c, d);
		if ((PEAK > 0)&&(peak == level)) level++;
	}
	if ((drawpeak == 1)&&(level >= oldlevel)&&(level < oldpeak)) {
		if (ORIENTATION == 0) {	b = POSY + HEIGHT - oldpeak; d = 1; }
		else if (ORIENTATION == 1) { b = POSY + oldpeak; d = 1; }
		else if (ORIENTATION == 2) { a = POSX + oldpeak; c = 1; }
		else	{ a = POSX + HEIGHT - oldpeak; c = 1; }
		XClearArea(display, rootWin, a, b, c, d, False);
	}
	if (level < oldlevel) {
		/* do me the pleasure and erase both at once */
		if (PEAK > 0) {
			if (oldlevel < oldpeak) oldlevel = oldpeak;
			if (oldlevel == peak) oldlevel--;
		}
				
		if (ORIENTATION == 0) { b = POSY + HEIGHT - oldlevel; d = oldlevel - level; }
		else if (ORIENTATION == 1) { b = POSY + level; d = oldlevel - level; }
		else if (ORIENTATION == 2) { a = POSX + level; c = oldlevel - level; }
		else	{ a = POSX + HEIGHT - oldlevel; c = oldlevel - level; }
		XClearArea(display, rootWin, a, b, c+SHADOW, d+SHADOW, False);
	}
	if ((drawpeak == 1)&&(peak > 0)) {
		XSetForeground(display, gc, peakcolor);
		if (ORIENTATION == 0) {	b = POSY + HEIGHT - peak; d = 1; }
		else if (ORIENTATION == 1) { b = POSY + peak; d = 1; }
		else if (ORIENTATION == 2) { a = POSX + peak; c = 1; }
		else	{ a = POSX + HEIGHT - peak; c = 1; }
		XFillRectangle(display, rootWin, gc, a, b, c, d);
	}
}

// Our worker thread

void* worker_func() {
	long pixel[6];
	gint16 freq_data[2][256];
	double scale, x00, y00;
	unsigned int i, j, level;
	unsigned short *level1, *level2, *levelsw, *peak1, *peak2, *peakstep;
	
	level1 = (unsigned short*)calloc(256, sizeof(short)); // need to be zeroed out
	level2 = (unsigned short*)malloc(256*sizeof(short));
	peak1 = (unsigned short*)calloc(256, sizeof(short)); // need to be zeroed out
	peak2 = (unsigned short*)malloc(256*sizeof(short));
	peakstep = (unsigned short*)calloc(256, sizeof(short)); // need to be zeroed out
	if ((level1 == NULL)||(level2 == NULL)||(peak1 == NULL)||(peak2 == NULL)||(peakstep == NULL))
		error_exit("Allocation of memory failed");
	
	initialize_X();
	pixel[0] = get_color(RED0, GREEN0, BLUE0);
	pixel[1] = get_color(RED1, GREEN1, BLUE1);
	pixel[2] = get_color(RED2, GREEN2, BLUE2);
	pixel[3] = get_color(RED3, GREEN3, BLUE3);
	pixel[4] = get_color(REDP, GREENP, BLUEP);
	pixel[5] = get_color(REDS, GREENS, BLUES);
	print_status("Ready to draw");
	
	scale = HEIGHT / (log((1 - LINEARITY) / LINEARITY) * 4);
        x00 = LINEARITY*LINEARITY*32768.0 / (2*LINEARITY - 1);
        y00 = -log(-x00) * scale;

	while (thread_control == GO) {
		pthread_mutex_lock(&mutex1);
		memcpy(&freq_data, &shared_freq_data, sizeof(gint16)*2*256);
		pthread_mutex_unlock(&mutex1);

		/* instead of copying the old level array to the second array, 
		we just tell the first is now the second one */
		levelsw = level1;
		level1 = level2;
		level2 = levelsw;
		levelsw = peak1;
		peak1 = peak2;
		peak2 = levelsw;
	
		for (i = 0; i < CUTOFF/DIV; i++) {
			level = 0;
			for (j = i*DIV; j < (i+1)*DIV; j++)
				if (level < freq_data[0][j])
					level = freq_data[0][j];
			level = level * (i*DIV + 1);
			level = floor(log(level - x00)*scale + y00);
			if (level < HEIGHT) {
				if ((level2[i] > FALLOFF)&&(level < level2[i] - FALLOFF))
					level1[i] = level2[i] - FALLOFF;
				else	level1[i] = level;
			} else level1[i] = HEIGHT;
			if (PEAK > 0) {
				if (level1[i] > peak2[i] - PEAK) {
					peak1[i] = level1[i];
					peakstep[i] = 0;
				} else if (peakstep[i] == PEAKSTEP)
					if (peak2[i] > PEAK) peak1[i] = peak2[i] - PEAK;
					else peak1[i] = 0;
				else {
					peak1[i] = peak2[i];
					peakstep[i]++;
				}
			}
			draw_bar(i, pixel[(int)ceil((float)level1[i] / HEIGHT * 4) - 1], pixel[4], pixel[5], level1[i], level2[i], peak1[i], peak2[i]);
		}
		XFlush(display);
		xmms_usleep(1000000 / FPS);
	}
	print_status("Worker thread: Exiting");
	XClearWindow(display, rootWin);
	XCloseDisplay(display); 
	free(level1);
	free(level2);
	return NULL;
}


// da xmms functions

static void rootvis_init(void) {
	int rc1;
	print_status("Initializing");
	
	clean_data();		
	thread_control = GO;
	if ((rc1 = pthread_create(&work_thread, NULL, worker_func, NULL))) {
		fprintf(stderr, "Thread creation failed: %d\n", rc1);
		error_exit("Thread creation failed");
	}
}

static void rootvis_cleanup(void) {
	print_status("Cleanup... ");
	thread_control = STOP;
	pthread_join(work_thread, NULL);
	print_status("Clean Exit");
}

static void rootvis_about(void)
{
	print_status("About");
}

static void rootvis_configure(void)
{
	print_status("Configure");
}

static void rootvis_playback_start(void)
{
	print_status("Playback starting");
}

static void rootvis_playback_stop(void)
{
	clean_data();
}

static void rootvis_render_freq(gint16 freq_data[2][256]) {
	int channel, bucket;

	pthread_mutex_lock(&mutex1);
	for (channel = 0; channel < 2; channel++) {
	 for (bucket = 0; bucket < 256; bucket++) {
		shared_freq_data[channel][bucket] = freq_data[channel][bucket];
	 }
	}
	pthread_mutex_unlock(&mutex1);
}
